import asyncio
import contextlib
import logging
from typing import TYPE_CHECKING

from .exceptions import FastMQTTError
from .structures import Message
from .subscription_manager import Subscription

if TYPE_CHECKING:
    from .fastmqtt import FastMQTT

log = logging.getLogger(__name__)


class MessageHandler:
    def __init__(self, fastmqtt: "FastMQTT"):
        self.fastmqtt = fastmqtt
        self._messages_handler_lock = asyncio.Lock()
        self._messages_handler_task = None

    async def __aenter__(self):
        self._messages_handler_task = asyncio.create_task(self._messages_handler())
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if self._messages_handler_task is None:
            return

        self._messages_handler_task.cancel()

        with contextlib.suppress(asyncio.CancelledError):
            await self._messages_handler_task

        self._messages_handler_task = None

    async def _wrap_message(
        self,
        subscription: Subscription,
        message: Message,
    ) -> None:
        results = await asyncio.gather(*(callback(message) for callback in subscription.callbacks))
        response_topic = message.properties.get("ResponseTopic")

        response_properties = None
        if "CorrelationData" in message.properties:
            response_properties = {"CorrelationData": message.properties["CorrelationData"]}

        for result in results:
            if result is not None and response_topic is None:
                raise FastMQTTError("Callback returned result, but message has no response_topic")

            if response_topic is not None:
                await self.fastmqtt.publish(
                    response_topic,
                    result,
                    properties=response_properties,
                )

    async def _messages_handler(self) -> None:
        if self._messages_handler_lock.locked():
            raise FastMQTTError("Messages handler is already running")

        async with self._messages_handler_lock:
            while True:
                try:
                    async for message in self.fastmqtt.client.messages:
                        message = Message._from_aiomqtt_message(self.fastmqtt, message)
                        if "SubscriptionIdentifier" not in message.properties:
                            log.warning(
                                "Message has no SubscriptionIdentifier (%s)",
                                message.topic,
                            )
                            continue

                        identifiers = message.properties["SubscriptionIdentifier"]

                        for identifier in identifiers:
                            subscription = self.fastmqtt._subscriptions_map.get(identifier)

                            if subscription is None:
                                log.error(
                                    "Message has unknown SubscriptionIdentifier %s (%s)",
                                    identifier,
                                    message.topic,
                                )
                                continue

                            asyncio.create_task(self._wrap_message(subscription, message))

                except asyncio.CancelledError:
                    return

                except Exception as e:
                    log.exception("Error in messages handler: %s", e)



import asyncio
import itertools
from typing import TYPE_CHECKING, Any, Callable

from aiomqtt.types import PayloadType

from .exceptions import FastMQTTError
from .structures import Message
from .subscription_manager import Retain, Subscription

if TYPE_CHECKING:
    from .fastmqtt import FastMQTT


class CorrelationIntGenerator:
    def __init__(self, limit: int = 2**16):
        self._correlation_data_counter = itertools.cycle(range(1, limit))

    def __call__(self) -> bytes:
        val = next(self._correlation_data_counter)
        return val.to_bytes((val.bit_length() + 7) // 8, "big")


class ResponseContext:
    def __init__(
        self,
        fastmqtt: "FastMQTT",
        response_topic: str,
        qos: int = 0,
        default_timeout: float | None = 60,
        correlation_generator: Callable[[], bytes] = CorrelationIntGenerator(),
    ):
        self._fastmqtt = fastmqtt
        self._response_topic = response_topic
        self._qos = qos
        self._default_timeout = default_timeout
        self._futures: dict[bytes, asyncio.Future[Message]] = {}
        self._subscription: Subscription | None = None
        self._identifier: int | None = None
        self._correlation_generator = correlation_generator

    async def subscribe(self) -> None:
        self._subscription = self._fastmqtt.register(
            self._callback,
            self._response_topic,
            self._qos,
            retain_handling=Retain.DO_NOT_SEND,
        )
        self._identifier = await self._fastmqtt._sub_manager.subscribe(self._subscription)

    async def close(self) -> None:
        if self._identifier is not None:
            await self._fastmqtt._sub_manager.unsubscribe(self._identifier, self._callback)

        self._identifier = None
        self._subscription = None
        for future in self._futures.values():
            future.cancel()

    async def __aenter__(self):
        await self.subscribe()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    async def _callback(self, message: Message) -> None:
        correlation_data = message.properties.get("CorrelationData")
        if correlation_data is None:
            raise FastMQTTError(
                f"correlation_data is None in response callback ({message.topic})"
            )

        future = self._futures.pop(correlation_data)
        future.set_result(message)

    async def request(
        self,
        topic: str,
        payload: PayloadType = None,
        qos: int = 0,
        retain: bool = False,
        properties: dict[str, Any] | None = None,
    ) -> Message:
        if properties is None:
            properties = {}

        correlation_data = self._correlation_generator()

        properties["CorrelationData"] = correlation_data
        properties["ResponseTopic"] = self._response_topic

        if correlation_data in self._futures:
            raise FastMQTTError(f"correlation_data {correlation_data} already in use")

        future = asyncio.Future[Message]()
        self._futures[correlation_data] = future
        try:
            await self._fastmqtt.publish(
                topic=topic,
                payload=payload,
                qos=qos,
                retain=retain,
                properties=properties,
            )
            return await future

        finally:
            self._futures.pop(correlation_data, None)



import asyncio
import logging
import ssl
from typing import Any, Callable, Iterable, Sequence

import paho.mqtt.client as mqtt
from aiomqtt import Client as MQTTClient
from aiomqtt import (
    Message,
    ProtocolVersion,
    ProxySettings,
    TLSParameters,
    Will,
)
from aiomqtt.types import PayloadType, SocketOption
from paho.mqtt.packettypes import PacketTypes

from .message_handler import MessageHandler
from .response import ResponseContext
from .router import MQTTRouter
from .subscription_manager import Subscription, SubscriptionManager
from .utils import properties_from_dict

WebSocketHeaders = dict[str, str] | Callable[[dict[str, str]], dict[str, str]]


class FastMQTT(MQTTRouter):
    def __init__(
        self,
        hostname: str,
        port: int = 1883,
        *,
        username: str | None = None,
        password: str | None = None,
        logger: logging.Logger | None = None,
        identifier: str | None = None,
        queue_type: type[asyncio.Queue[Message]] | None = None,
        will: Will | None = None,
        clean_session: bool | None = None,
        transport: str = "tcp",
        timeout: float | None = None,
        keepalive: int = 60,
        bind_address: str = "",
        bind_port: int = 0,
        clean_start: int = mqtt.MQTT_CLEAN_START_FIRST_ONLY,
        max_queued_incoming_messages: int | None = None,
        max_queued_outgoing_messages: int | None = None,
        max_inflight_messages: int | None = None,
        max_concurrent_outgoing_calls: int | None = None,
        properties: mqtt.Properties | None = None,
        tls_context: ssl.SSLContext | None = None,
        tls_params: TLSParameters | None = None,
        tls_insecure: bool | None = None,
        proxy: ProxySettings | None = None,
        socket_options: Iterable[SocketOption] | None = None,
        websocket_path: str | None = None,
        websocket_headers: WebSocketHeaders | None = None,
        routers: Sequence[MQTTRouter] | None = None,
    ):
        super().__init__()
        self.client = MQTTClient(
            hostname,
            port,
            username=username,
            password=password,
            logger=logger,
            identifier=identifier,
            queue_type=queue_type,
            protocol=ProtocolVersion.V5,
            will=will,
            clean_session=clean_session,
            transport=transport,
            timeout=timeout,
            keepalive=keepalive,
            bind_address=bind_address,
            bind_port=bind_port,
            clean_start=clean_start,
            max_queued_incoming_messages=max_queued_incoming_messages,
            max_queued_outgoing_messages=max_queued_outgoing_messages,
            max_inflight_messages=max_inflight_messages,
            max_concurrent_outgoing_calls=max_concurrent_outgoing_calls,
            properties=properties,
            tls_context=tls_context,
            tls_params=tls_params,
            tls_insecure=tls_insecure,
            proxy=proxy,
            socket_options=socket_options,
            websocket_path=websocket_path,
            websocket_headers=websocket_headers,
        )
        self._subscriptions_map: dict[int, Subscription] = {}
        self._sub_manager = SubscriptionManager(self)
        self._message_handler = MessageHandler(self)
        self._state: dict[str, Any] = {}

        if routers is not None:
            for router in routers:
                self.include_router(router)

    @property
    def identifier(self) -> str:
        return self.client.identifier

    def __setitem__(self, key: str, value: Any) -> None:
        self._state[key] = value

    def __getitem__(self, key: str) -> Any:
        return self._state[key]

    def __delitem__(self, key: str) -> None:
        del self._state[key]

    async def __aenter__(self):
        self._started = True
        await self.client.__aenter__()
        await self._message_handler.__aenter__()
        await self.subscribe_all()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.client.__aexit__(exc_type, exc_value, traceback)
        await self._message_handler.__aexit__(exc_type, exc_value, traceback)

    async def subscribe_all(self) -> None:
        await self._sub_manager.subscribe_all()

    async def publish(
        self,
        topic: str,
        payload: PayloadType = None,
        qos: int = 0,
        retain: bool = False,
        properties: dict[str, Any] | None = None,
    ) -> None:
        mqtt_properties = None
        if properties is not None:
            mqtt_properties = properties_from_dict(properties, PacketTypes.PUBLISH)

        await self.client.publish(
            topic=topic,
            payload=payload,
            qos=qos,
            retain=retain,
            properties=mqtt_properties,
        )

    def response_context(
        self,
        response_topic: str,
        qos: int = 0,
        default_timeout: float | None = 60,
        **kwargs,
    ) -> ResponseContext:
        return ResponseContext(
            self,
            response_topic,
            qos,
            default_timeout,
            **kwargs,
        )



import logging
from typing import Any, Callable

import aiomqtt

from .exceptions import FastMQTTError
from .subscription_manager import (
    CallbackType,
    Retain,
    Subscription,
)

log = logging.getLogger(__name__)


class MQTTRouter:
    def __init__(self):
        self._subscriptions: list[Subscription] = []
        self._started = False

    def _check_different(self, subscription: Subscription, **new_attrs) -> None:
        for name, value in new_attrs.items():
            exist_sub_attr = getattr(subscription, name)

            if exist_sub_attr != value:
                log.warning(
                    "Subscription %s has different %s. Existing: %s, New: %s",
                    subscription.topic,
                    name,
                    exist_sub_attr,
                    value,
                )

    def include_router(self, router: "MQTTRouter") -> None:
        if self._started:
            raise FastMQTTError("Cannot include router after starting")

        included_subscriptions = self._subscriptions.copy()
        for router_sub in router._subscriptions:
            for included_sub in included_subscriptions:
                if included_sub.topic == router_sub.topic:
                    included_sub.callbacks.extend(router_sub.callbacks)
                    break
            else:
                self._subscriptions.append(router_sub)

    def register(
        self,
        callback: CallbackType,
        topic: str,
        qos: int = 0,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_handling: Retain = Retain.SEND_ON_SUBSCRIBE,
    ) -> Subscription:
        for subscription in self._subscriptions:
            if str(subscription.topic) == topic:
                subscription.callbacks.append(callback)
                self._check_different(
                    subscription,
                    qos=qos,
                    no_local=no_local,
                    retain_as_published=retain_as_published,
                    retain_handling=retain_handling,
                )
                return subscription

        subscription = Subscription(
            [callback],
            aiomqtt.Topic(topic),
            qos,
            no_local,
            retain_as_published,
            retain_handling,
        )

        self._subscriptions.append(subscription)

        return subscription

    def on_message(
        self,
        topic: str,
        qos: int = 0,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_handling: Retain = Retain.SEND_ON_SUBSCRIBE,
    ) -> Callable[..., Any]:
        def wrapper(callback: CallbackType) -> CallbackType:
            self.register(
                callback,
                topic,
                qos,
                no_local,
                retain_as_published,
                retain_handling,
            )
            return callback

        return wrapper



import itertools
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Awaitable, Callable

import aiomqtt
from paho.mqtt.packettypes import PacketTypes
from paho.mqtt.subscribeoptions import SubscribeOptions

from .exceptions import FastMQTTError
from .structures import Message
from .utils import properties_from_dict

if TYPE_CHECKING:
    from .fastmqtt import FastMQTT


CallbackType = Callable[[Message], Awaitable[Any]]


class Retain(IntEnum):
    SEND_ON_SUBSCRIBE = 0
    SEND_IF_NEW_SUB = 1
    DO_NOT_SEND = 2


@dataclass
class Subscription:
    callbacks: list[CallbackType]
    topic: aiomqtt.Topic
    qos: int = 0
    no_local: bool = False
    retain_as_published: bool = False
    retain_handling: Retain = Retain.SEND_ON_SUBSCRIBE


class SubscriptionManager:
    def __init__(self, fastmqtt: "FastMQTT"):
        self.fastmqtt = fastmqtt
        self.subscription_id_counter = itertools.count(1)

    async def subscribe(self, subscription: Subscription) -> int:
        identifier = next(self.subscription_id_counter)
        properties = properties_from_dict(
            {"SubscriptionIdentifier": identifier},
            PacketTypes.SUBSCRIBE,
        )

        await self.fastmqtt.client.subscribe(
            topic=subscription.topic.value,
            qos=subscription.qos,
            options=SubscribeOptions(
                qos=subscription.qos,
                noLocal=subscription.no_local,
                retainAsPublished=subscription.retain_as_published,
                retainHandling=subscription.retain_handling,
            ),
            properties=properties,
        )
        self.fastmqtt._subscriptions_map[identifier] = subscription
        return identifier

    async def subscribe_all(self) -> None:
        for subscription in self.fastmqtt._subscriptions:
            await self.subscribe(subscription)

    async def unsubscribe(self, identifier: int, callback: CallbackType | None = None) -> None:
        subscription = self.fastmqtt._subscriptions_map.get(identifier)
        if subscription is None:
            raise FastMQTTError(f"Unknown subscription identifier {identifier}")

        if callback is not None:
            subscription.callbacks.remove(callback)

        if callback is None or not subscription.callbacks:
            await self.fastmqtt.client.unsubscribe(subscription.topic.value)
            del self.fastmqtt._subscriptions_map[identifier]



from typing import Any

from paho.mqtt.properties import Properties


def properties_from_dict(properties: dict[str, Any], packet_type: int) -> Properties:
    mqtt_properties = Properties(packet_type)
    for key, value in properties.items():
        setattr(mqtt_properties, key, value)

    return mqtt_properties


def properties_to_dict(properties: Properties) -> dict[str, Any]:
    data = {}
    for name in properties.names:
        compressed_name = name.replace(" ", "")
        if hasattr(properties, compressed_name):
            val = getattr(properties, compressed_name)
            data[compressed_name] = val
    return data



from .exceptions import FastMQTTError
from .fastmqtt import FastMQTT
from .router import MQTTRouter
from .structures import Message
from .subscription_manager import Retain, Subscription

__all__ = [
    "FastMQTT",
    "FastMQTTError",
    "Message",
    "MQTTRouter",
    "Retain",
    "Subscription",
]



import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import aiomqtt

from .utils import properties_to_dict

if TYPE_CHECKING:
    from .fastmqtt import FastMQTT


class Payload:
    def __init__(self, data: bytes) -> None:
        self._data = data

    def raw(self) -> bytes:
        return self._data

    def text(self) -> str:
        return self._data.decode("utf-8")

    def json(self, **kwargs) -> dict[str, Any]:
        return json.loads(self._data, **kwargs)


@dataclass(frozen=True)
class Message:
    fastmqtt: "FastMQTT"

    topic: str
    payload: Payload
    qos: int
    retain: bool
    mid: int
    properties: dict[str, Any]

    @classmethod
    def _from_aiomqtt_message(cls, fastmqtt: "FastMQTT", message: aiomqtt.Message) -> "Message":
        return cls(
            fastmqtt=fastmqtt,
            topic=str(message.topic),
            payload=Payload(message.payload),  # type: ignore
            qos=message.qos,
            retain=message.retain,
            mid=message.mid,
            properties=properties_to_dict(message.properties)
            if message.properties is not None
            else {},
        )
